""" 
Place to define providers.
Most part of _BASE_PROVIDERS was taken from https://github.com/vincecarney/dnsbl
"""

import requests
from bs4 import BeautifulSoup

from dnsbl_checker.utils import parse_file

# providers answers could be interpreted in one of the following categories
DNSBL_CATEGORIES = {'spam', 'proxy', 'malware', 'botnet', 'exploits', 'unknown'}


class Provider(object):

    def __init__(self, host):
        self.host = host

    def process_response(self, response):
        result = set()
        if response:
            return {'unknown'}
        return result

    def __repr__(self):
        return "<Provider: %s>" % self.host


class ZenSpamhaus(Provider):
    """ Combined spamhaus list:
        https://www.spamhaus.org/faq/section/DNSBL%20Usage#200
    """

    def __init__(self, host='zen.spamhaus.org'):
        Provider.__init__(self, host=host)

    def process_response(self, response):
        categories = set()
        for result in response:
            if result.host in ['127.0.0.2', '127.0.0.3', '127.0.0.9']:
                categories.add('spam')
            if result.host in ['127.0.0.4', '127.0.0.5', '127.0.0.6',
                               '127.0.0.7']:
                categories.add('exploits')
        return categories


def update_providers(fname, banned_providers=None):
    print("Getting list of black lists...")

    try:
        resp = requests.get("http://multirbl.valli.org/list/")
    except requests.exceptions.RequestException:
        print("Failed to fetch remote black lists. Continue...")
        return

    if resp.status_code != 200:
        print("Failed to fetch remote black lists. Continue...")
        return

    print("Received black lists")

    soup = BeautifulSoup(resp.content, 'html.parser')

    # rewrite providers
    with open(fname, 'w+') as file:
        for row in soup.find("table").find_all('tr'):
            if row.contents[6].next != 'b':  # provider is blacklist
                continue

            black_list = row.contents[2].next
            if (black_list == "(hidden)"
                    or black_list in banned_providers):
                continue

            file.write(black_list + '\n')


def get_providers(fname):
    providers = [ZenSpamhaus()]

    for host in parse_file(fname) or list():
        providers.append(Provider(host))

    return providers



import os


def parse_file(file):
    if file is None or not os.path.isfile(file):
        return list()

    items = list()
    with open(file, "rt") as list_file:
        for line in list_file:
            item = line.split('#', 1)[0].rstrip()
            if item:
                items.append(item)
    return items



import copy


class Ip:
    def __init__(self, db, addr):
        self.db = db
        self.addr = addr

        self.blacklists = self.db.get(addr, dict())

    def save(self):
        self.db[self.addr] = self.blacklists


class Saver:
    def __init__(self, db):
        self._db = db

        self.changes = {}

    def save_results(self, results):
        for result in results:
            ip = Ip(self._db, result.addr)

            change_data = dict()

            new_blacklists = set(result.detected_by) - set(ip.blacklists)
            for bl in new_blacklists:
                data = {bl: result.detected_by[bl]}
                change_data.setdefault('added_bls', dict()).update(data)

                ip.blacklists.update(data)

            blacklists_to_delete = set(ip.blacklists) - set(result.detected_by)
            for bl in blacklists_to_delete:
                if bl in result.failed_providers:
                    continue

                change_data.setdefault('deleted_bls', dict()).update({
                    bl: copy.deepcopy(ip.blacklists[bl])
                })

                del ip.blacklists[bl]

            if change_data:
                self.changes[ip.addr] = change_data

            ip.save()

    def delete_providers(self, providers):
        for provider in providers:
            for ip, bls in self._db.items():
                if provider in bls:
                    del self._db[ip][provider]
                    break



import socket
import asyncio
import aiodns

from dnsbl_checker.providers import DNSBL_CATEGORIES


class DNSBLResult(object):
    """
    DNSBL Result class to keep all info about ip request results.

    """
    def __init__(self, addr=None, results=None):
        self.addr = addr
        self._results = results
        self.blacklisted = False
        self.providers = []
        self.failed_providers = []
        self.detected_by = {}
        self.categories = set()
        self.process_results()

    def process_results(self):
        """ Process results by providers """
        for result in self._results:
            provider = result.provider
            self.providers.append(provider)
            if result.error:
                self.failed_providers.append(provider.host)
                continue

            if not result.response['a_response']:
                continue

            # set blacklisted to True if ip is detected with at least one dnsbl
            self.blacklisted = True

            provider_categories = provider.process_response(
                result.response['a_response']
            )
            assert provider_categories.issubset(DNSBL_CATEGORIES)
            self.categories = self.categories.union(provider_categories)

            txt_response = result.response['txt_response']

            self.detected_by[str(provider.host)] = dict(
                result=dict(
                    a_response=result.response['a_response'][0].host,
                    txt_response=(txt_response[0].text.decode('utf-8')
                                  if txt_response else "TXT request failed")
                ),
                catogories=list(provider_categories)
            )

    def __repr__(self):
        blacklisted = '[BLACKLISTED]' if self.blacklisted else ''
        return "<DNSBLResult: %s %s (%d/%d)>" % (self.addr, blacklisted,
                                                 len(self.detected_by),
                                                 len(self.providers))


class DNSBLResponse(object):
    """
    DNSBL Response object
    """
    def __init__(self, addr=None, provider=None, response=None, error=None):
        self.addr = addr
        self.provider = provider
        self.response = response
        self.error = error


class DNSBLChecker(object):
    def __init__(self,
                 providers,
                 timeout=5,
                 tries=2,
                 concurrency=200,
                 loop=None):

        self.providers = providers

        self.progress = None

        self._loop = loop or asyncio.get_event_loop()
        self._resolver = aiodns.DNSResolver(
            timeout=timeout,
            tries=tries,
            loop=self._loop
        )
        self._semaphore = asyncio.Semaphore(concurrency, loop=self._loop)

    async def dnsbl_request(self, addr, provider):
        """
        Make lookup to dnsbl provider
        Parameters:
            * addr (string) - ip address to check
            * provider (string) - dnsbl provider

        Returns:
            * DNSBLResponse object

        Raises:
            * ValueError
        """
        ip_reversed = '.'.join(reversed(addr.split('.')))
        dnsbl_query = "%s.%s" % (ip_reversed, provider.host)

        a_response = None
        txt_response = None
        error = None

        async with self._semaphore:
            try:
                a_response = await self._resolver.query(dnsbl_query, 'A')
            except aiodns.error.DNSError as exc:
                if exc.args[0] != 4:  # 4: domain name not found:
                    error = exc
            else:
                try:
                    txt_response = await self._resolver.query(
                        dnsbl_query, 'TXT'
                    )
                except aiodns.error.DNSError:
                    pass

        if self.progress is not None:
            self.progress.update(1)

        return DNSBLResponse(
            addr=addr,
            provider=provider,
            response=dict(a_response=a_response, txt_response=txt_response),
            error=error
        )

    async def _check_ip(self, addr):
        try:
            socket.inet_aton(addr)
        except socket.error:
            raise ValueError('wrong ip format')

        tasks = []
        for provider in self.providers:
            tasks.append(self.dnsbl_request(addr, provider))
        results = await asyncio.gather(*tasks)

        return DNSBLResult(addr=addr, results=results)

    def check_ips(self, addrs):
        tasks = []
        for addr in addrs:
            tasks.append(self._check_ip(addr))
        return self._loop.run_until_complete(asyncio.gather(*tasks))



import asyncio
import json
import shelve
import sys
import os
import errno

import click
import yaml
import tqdm

from dnsbl_checker.utils import parse_file
from dnsbl_checker.providers import update_providers, get_providers
from dnsbl_checker.checker import DNSBLChecker
from dnsbl_checker.saver import Saver
from dnsbl_checker.telegram import TGClient

dir_path = os.path.dirname(os.path.realpath(__file__))

data_dir = os.path.join(dir_path, 'data')

db_file = os.path.join(data_dir, 'blacklisted')
providers_file = os.path.join(data_dir, 'providers.txt')

if not os.path.exists(data_dir):
    try:
        os.makedirs(data_dir)
    except OSError as exc:  # Guard against race condition
        if exc.errno != errno.EEXIST:
            raise

    with open(providers_file, 'w+'):
        pass

    with shelve.open(db_file, flag='n'):
        pass


@click.group()
def cli():
    pass


@cli.command()
@click.argument('ip_addr')
def get(ip_addr):
    with shelve.open(db_file, flag='r') as db:
        if ip_addr == 'all':
            data = db
        else:
            try:
                data = db[ip_addr]
            except KeyError:
                print('IP {} was not found'.format(ip_addr))
                sys.exit(1)

        print(json.dumps(dict(data), indent=2, sort_keys=True))


@cli.command()
@click.argument('conf_file', type=click.File(mode='r'))
def check(conf_file):
    config = yaml.load(conf_file)

    if 'ips' not in config:
        print('"ips" param is required')

    loop = asyncio.get_event_loop()

    banned_providers = parse_file(config.get('banned_providers'))

    result = inspect(
        parse_file(config['ips']),
        loop,
        banned_providers
    )

    with shelve.open(db_file) as db:
        saver = Saver(db)

        if banned_providers:
            saver.delete_providers(banned_providers)

        saver.save_results(result)

    if 'telegram_token' in config and 'telegram_ids' in config:
        print('Sending notifications...')

        notify(config['telegram_token'],
               config['telegram_ids'],
               loop,
               saver.changes)

        print('Notifications were sent')
    else:
        print('Skipping notifications')


def inspect(ips, loop, banned_providers=None):
    update_providers(providers_file,
                     banned_providers=banned_providers)

    checker = DNSBLChecker(get_providers(providers_file),
                           loop=loop,
                           concurrency=300)

    with tqdm.tqdm(total=len(ips) * len(checker.providers),
                   miniters=1,
                   smoothing=0,
                   desc="IP checks",
                   unit=" requests") as checker.progress:
        return checker.check_ips(ips)


def notify(bot_token, chat_ids, loop, data=None):
    tg_client = TGClient(
        bot_token,
        chat_ids,
        loop=loop
    )

    if data:
        msgs = prepare_tg_msgs(data)
        tg_client.send_msgs(*msgs)
    else:
        tg_client.send_msgs('Changes were not found')


def prepare_tg_msgs(changes):
    msgs = list()

    def get_record(bl, data):
        resp = data['result']

        return (
            '*{host}*: {a_addr}\nTXT response: {txt}'.format(
                host=bl,
                a_addr=resp['a_response'],
                txt=resp['txt_response']
            )
        )

    for ip, change in changes.items():

        if change.get('added_bls'):
            msg = '{}\nFound in:\n{}'
            for bl, data in change['added_bls'].items():
                msgs.append(msg.format(ip, get_record(bl, data)))

        if change.get('deleted_bls'):
            msg = '{}\nDeleted from:\n{}'
            for bl, data in change['deleted_bls'].items():
                msgs.append(msg.format(ip, get_record(bl, data)))

    return msgs



#!/usr/bin/env python

from os import path as op

from setuptools import setup


def _read(fname):
    try:
        return open(op.join(op.dirname(__file__), fname)).read()
    except IOError:
        return ''


install_requires = [
    l for l in _read('requirements.txt').split('\n')
    if l and not l.startswith('#')
]

setup(
    name='dnsbl_checker',
    version='0.0.2',
    description='Asynchronous script for monitoring dns blacklists',
    long_description=_read('README.md'),
    author='Vadim Zhyvylo',
    author_email='jyvylo5@gmail.com',
    url='https://github.com/VadimZhyvyloKS/dnsbl-checker',
    license='MIT',
    install_requires=install_requires,
    keywords='dns blacklist dnsbl async asynchronous check',
    include_package_data=True,
    packages=['dnsbl_checker'],
    entry_points='''
        [console_scripts]
        dnsbl=dnsbl_checker.cli:cli
    '''
)



import asyncio

import aiohttp


class TGClient:
    def __init__(self, token, chat_ids, loop=None):
        self.url = 'https://api.telegram.org/bot{}/sendMessage'.format(token)
        self.chat_ids = chat_ids

        self._loop = loop or asyncio.get_event_loop()
        self._session = lambda: aiohttp.ClientSession(loop=self._loop)
        self._semaphore = asyncio.Semaphore(5, loop=self._loop)

    async def _send_request(self, client, data):
        async with self._semaphore:
            async with client.post(self.url, data=data) as response:
                return await response.text()

    async def _send_msg(self, msg, client):
        tasks = []
        for chat_id in self.chat_ids:
            data = dict(
                chat_id=chat_id,
                text=msg,
                parse_mode='Markdown',
                disable_web_page_preview=True
            )

            tasks.append(self._send_request(client, data))

        return await asyncio.gather(*tasks)

    async def async_send_msgs(self, msgs):
        tasks = []
        async with self._session() as client:
            for msg in msgs:
                tasks.append(self._send_msg(msg, client))

            return await asyncio.gather(*tasks)

    def send_msgs(self, *msgs):
        return self._loop.run_until_complete(self.async_send_msgs(msgs))
